/*
 *  GuiLoader, a GuiXml loader library
 *  Copyright (c) 2006-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#include "guiloader_impl.h"
#include "proxies.h"

GUILOADER_API int guiloader_version(void) {
	return GUILOADER_VERSION;
}

GUILOADER_API void guiloader_enter_check(const char * msg) {
	GuiLoaderClass * meta= GUI_LOADER_CLASS(g_type_class_ref(GUI_TYPE_LOADER));
	InitGettext();
	if(!meta->init)
		meta->init= NewInit();
	else
		++meta->init->refcounter;
	EnterCounter(meta->init, msg);
	g_type_class_unref(meta);
}

GUILOADER_API void guiloader_leave(void) {
	GuiLoaderClass * meta= GUI_LOADER_CLASS(g_type_class_ref(GUI_TYPE_LOADER));
	LeaveCounter(meta->init);
	--meta->init->refcounter;
	if(meta->init->refcounter == 0) {
		FreeInit(meta->init);
		meta->init= NULL;
	}
	g_type_class_unref(meta);
}

GUILOADER_API GType gui_loader_get_type(void) {
	static GType object_type= 0;
	if(object_type==0) {
		static const GTypeInfo object_info= {
			sizeof(GuiLoaderClass),
			NULL,
			NULL,
			(GClassInitFunc)gui_loader_class_init,
			NULL,
			NULL,
			sizeof(GuiLoader),
			0,
			(GInstanceInitFunc)gui_loader_init,
		};
	  object_type= g_type_register_static(G_TYPE_OBJECT, "GuiLoader", &object_info, 0);
	}
	return object_type;
}

GuiLoader * gui_loader_new(void) {
	return g_object_new(GUI_TYPE_LOADER, NULL);
}

GUILOADER_API void gui_loader_set_preview_mode(GuiLoader * loader, gboolean preview) {
	GuiLoaderPrivate * priv;
	CHECK(GUI_IS_LOADER(loader));
	priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->preview= preview;
	CHECK(!priv->preview || !priv->take);
}

GUILOADER_API void gui_loader_set_take_objects(GuiLoader * loader, gboolean take) {
	GuiLoaderPrivate * priv;
	CHECK(GUI_IS_LOADER(loader));
	priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->take= take;
	CHECK(!priv->preview || !priv->take);
}

GUILOADER_API void gui_loader_add_custom_constructor(GuiLoader * loader, const char * type, GuiCustomConstructorFunc func, gpointer data, GDestroyNotify destroy) {
	GuiLoaderPrivate * priv;
	CustomConstructor * constructor;
	CHECK(GUI_IS_LOADER(loader) && type && func);
	priv= GUI_LOADER_GET_PRIVATE(loader);
	constructor= g_slice_new(CustomConstructor);
	constructor->func= func;
	constructor->data= data;
	constructor->destroy= destroy;
	g_hash_table_insert(priv->constructors, g_strdup(type), constructor);
}

GUILOADER_API void gui_loader_load_from_string(GuiLoader * loader, const char * data, int size) {
	CHECK(GUI_IS_LOADER(loader) && data);
	FreeGui(loader);
	LoadModel(loader, data, size);
	LoadGui(loader);
	FreeModel(loader);
}

GUILOADER_API gboolean gui_loader_load_from_file(GuiLoader * loader, const char * filename, GError ** error) {
	char * data;
	gsize size;
	gboolean result;
	CHECK(GUI_IS_LOADER(loader));
	result= g_file_get_contents(filename, &data, &size, error);
	if(result) {
		gui_loader_load_from_string(loader, data, size);
		g_free(data);
	}
	return result;
}

GUILOADER_API void gui_loader_clear(GuiLoader * loader) {
	CHECK(GUI_IS_LOADER(loader));
	FreeGui(loader);
}

GUILOADER_API GQueue * gui_loader_query_connectors(GuiLoader * loader, const char * name, int count) {
	GuiLoaderPrivate * priv;
	GQueue * queue;
	CHECK(GUI_IS_LOADER(loader) && name);
	priv= GUI_LOADER_GET_PRIVATE(loader);
	queue= g_hash_table_lookup(priv->connectors, name);
	CHECK(queue && SIGNED(g_queue_get_length(queue))==count);
	return queue;
}

GUILOADER_API void gui_loader_get_connect_info(GuiLoader * loader, const GuiConnector * connector, GuiConnectInfo * info) {
	CHECK(GUI_IS_LOADER(loader) && connector && info);
	info->object= connector->object;
	info->signame= ShortStrGet(connector->emitter->signame);
	info->after= connector->emitter->after;
}

GUILOADER_API void gui_loader_connect_signals(GuiLoader * loader, const char * name, GCallback func, gpointer userdata, int count) {
	GQueue * queue= gui_loader_query_connectors(loader, name, count);
	GList * elem;
	for(elem= g_queue_peek_head_link(queue); elem; elem= g_list_next(elem)) {
		GuiConnectInfo info;
		gulong hid;
		gui_loader_get_connect_info(loader, elem->data, &info);
		hid= g_signal_connect_data(
			info.object,
			info.signame,
			func, userdata, NULL,
			info.after ? G_CONNECT_AFTER : 0
		);
		if(!hid)
			FatalError(_("Connecting signals failed (see log messages)"));
	}
}

GUILOADER_API gpointer gui_loader_bind_object(GuiLoader * loader, const char * name) {
	GObject * object;
	CHECK(GUI_IS_LOADER(loader) && name);
	object= FindObject(loader, name);
	CHECK(object);
	if(GTK_IS_WINDOW(object))
		g_object_set_data(object, "GuiLoaderBoundWindow", "true");
	return object;
}

void gui_loader_class_init(GuiLoaderClass * meta) {
	GObjectClass * object_class;
	meta->init= NULL;
	meta->parent_class= g_type_class_peek_parent(meta);
	object_class= G_OBJECT_CLASS(meta);
	object_class->finalize= gui_loader_finalize;
	g_type_class_add_private(object_class, sizeof(GuiLoaderPrivate));
}

void gui_loader_init(GuiLoader * loader) {
	InitInstance(loader);
}

void gui_loader_finalize(GObject * object) {
	GuiLoader * loader= GUI_LOADER(object);
	GuiLoaderClass * meta= GUI_LOADER_GET_CLASS(loader);
	DoneInstance(loader);
	(*meta->parent_class->finalize)(object);
}

void CheckFailed(const char * expr, const char * file, int line) {
	char msg[256];
	g_snprintf(msg, sizeof(msg), _("Check failed: %s, file: %s, line: %d"), expr, file, line);
	FatalError(msg);
}

void FatalError(const char * message) {
	char buf[128];
	GtkWidget * dialog;
	g_critical(message);

	#ifdef GUILOADER_SANDBOX
		assert(0);
	#endif

	g_snprintf(buf, sizeof(buf), _("Fatal Error [%s]"), G_LOG_DOMAIN);

	dialog= gtk_message_dialog_new(
		NULL,
		GTK_DIALOG_MODAL,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_CLOSE,
		buf
	);

	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), message);
	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	abort();
}

int NoOp(void) {
	return 0;
}

ShortStr * ShortStrNew(const char * str) {
	int len= strlen(str);
	return ShortStrNewLen(str, len);
}

ShortStr * ShortStrNewLen(const char * str, int len) {
	ShortStr * string= ShortStrAlloc(len);
	char * buf= ShortStrGet(string);
	strncpy(buf, str, len);
	buf[len]= 0;
	return string;
}

ShortStr * ShortStrAlloc(int len) {
	ShortStr * string;
	CHECK(len <= 65535);
	string= g_slice_alloc(sizeof(ShortStr) + len);
	string->refcounter= 1;
	string->len= len;
	return string;
}

char * ShortStrGet(ShortStr * string) {
	return string ? &string->buf[3] : NULL;
}

void ShortStrRef(ShortStr * string) {
	CHECK(string->refcounter < 255);
	++string->refcounter;
}

void ShortStrUnref(ShortStr * string) {
	--string->refcounter;
	if(string->refcounter==0)
		g_slice_free1(sizeof(ShortStr) + string->len, string);
}

ShortStr * ShortStrReplace(ShortStr * string, int pos, int len, const char * repstr) {
	ShortStr * result;
	char * sbuf, * rbuf;
	int replen;
	CHECK(SIGNED(string->len) >= pos+len);
	replen= strlen(repstr);
	result= ShortStrAlloc(string->len - len + replen);
	sbuf= ShortStrGet(string);
	rbuf= ShortStrGet(result);
	strncpy(rbuf, sbuf, pos);
	strncpy(rbuf+pos, repstr, replen);
	strncpy(rbuf+pos+replen, sbuf+pos+len, string->len - pos - len);
	rbuf[result->len]= 0;
	ShortStrUnref(string);
	return result;
}

guint ShortStrHash(gconstpointer string) {
	return g_str_hash(ShortStrGet((gpointer)string));
}

gboolean ShortStrEqual(gconstpointer s1, gconstpointer s2) {
	return g_str_equal(ShortStrGet((gpointer)s1), ShortStrGet((gpointer)s2));
}

gboolean ShortStrEmpty(ShortStr * string) {
	return string ? string->len==0 : TRUE;
}

PtrArray * PtrArrayNew(int size) {
	PtrArray * array= g_slice_alloc(sizeof(PtrArray) + sizeof(gpointer)*(size - 1));
	array->size= size;
	return array;
}

void PtrArrayFree(PtrArray * array) {
	if(array)
		g_slice_free1(sizeof(PtrArray) + sizeof(gpointer)*(array->size - 1), array);
}

void PtrArrayNull(PtrArray * array) {
	int i;
	for(i= 0; i<array->size; ++i)
		array->ptrs[i]= NULL;
}

IntArray * IntArrayNew(int size) {
	IntArray * array= g_slice_alloc(sizeof(IntArray) + sizeof(int)*(size - 1));
	array->size= size;
	return array;
}

void IntArrayFree(IntArray * array) {
	if(array)
		g_slice_free1(sizeof(IntArray) + sizeof(int)*(array->size - 1), array);
}

GList * CreateListFromPtrArray(PtrArray * array) {
	GList * result= NULL;
	int i;
	for(i= array->size - 1; i >= 0; --i)
		result= g_list_prepend(result, array->ptrs[i]);
	return result;
}

pconstchar * SplitString(const char * str, char ch, int * maxcount) {
	int count= 1;
	int length;
	pconstchar * ptrs;
	int index= 0;
	int i;
	for(length= 0; str[length]; ++length)
		if(str[length]==ch)
			++count;
	if(*maxcount > 0)
		count= MIN(count, *maxcount);
	ptrs= g_slice_alloc(sizeof(gpointer)*(count+1));
	ptrs[index++]= str;
	for(i= 0; i<length && index<count; ++i)
		if(str[i]==ch)
			ptrs[index++]= &str[i+1];
	ptrs[index++]= &str[length+1];
	CHECK(index==count+1);
	*maxcount= count;
	return ptrs;
}

int SplitLength(pconstchar * ptrs, int i) {
	return ptrs[i+1] - ptrs[i] - 1;
}

void FreeSplitPtrs(pconstchar * ptrs, int count) {
	g_slice_free1(sizeof(gpointer)*(count+1), (gpointer)ptrs);
}

gboolean BoolFromString(const char * str) {
	gboolean var= FALSE;
	if(StrEqual(str, "true"))
		var= TRUE;
	else if(StrEqual(str, "false"))
		var= FALSE;
	else
		CHECK(FALSE);
	return var;
}

const char * Translate(const char * meta, const char * str, int * size) {
	const char * result= str;

	if(meta && *size > 0) {

		int count= 3;
		pconstchar * vstr= SplitString(meta, '|', &count);
		const char * prefix= NULL;
		int prefix_size= 0;

		if(strncmp(vstr[0], "translate", 9) == 0) {
			if(count > 1) {
				prefix= vstr[1];
				prefix_size= SplitLength(vstr, 1);
			}

			if(prefix_size==0)
				result= gettext(str);
			else {
				int buf_size= prefix_size + 1 + *size + 1;
				char * buf= g_slice_alloc(buf_size);
				strncpy(buf, prefix, prefix_size);
				buf[prefix_size]= '|';
				strncpy(&buf[prefix_size+1], str, *size + 1);
				result= gettext(buf);
				g_slice_free1(buf_size, buf);
			}

			*size= strlen(result);
		}

		FreeSplitPtrs(vstr, count);
	}

	return result;
}

gboolean StrEqual(const char * s1, const char * s2) {
	return s1==s2 || (s1 && s2 && strcmp(s1, s2) == 0);
}

void InitInstance(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->preview= FALSE;
	priv->take= FALSE;
	priv->stack= NULL;
	priv->model= NULL;
	priv->version= 0;
	priv->objects= g_hash_table_new(g_str_hash, g_str_equal);
	priv->nested= NULL;
	priv->connectors= g_hash_table_new(g_str_hash, g_str_equal);
	priv->delayed= NULL;
	priv->strings= NULL;
	priv->constructors= g_hash_table_new_full(
		g_str_hash,
		g_str_equal,
		g_free,
		(GDestroyNotify)CustomConstructorFree
	);
}

void DoneInstance(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	FreeGui(loader);
	g_hash_table_destroy(priv->objects);
	g_hash_table_destroy(priv->connectors);
	g_hash_table_destroy(priv->constructors);
}

void CustomConstructorFree(CustomConstructor * constructor) {
	if(constructor->destroy)
		constructor->destroy(constructor->data);
	g_slice_free(CustomConstructor, constructor);
}

void OnStartElement(
	GMarkupParseContext * context,
	const char * element,
	const char ** attr_names,
	const char ** attr_values,
	gpointer user_data,
	GError ** error
) {

	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(user_data);

	Node * node= NewNode();

	if(StrEqual(element, "gui")) {

		int version= 0;
		const char * nspace= NULL;

		int i;
		for(i=0; attr_names[i]; ++i) {
			if(StrEqual(attr_names[i], "namespace"))
				nspace= attr_values[i];
			else if(StrEqual(attr_names[i], "version"))
				version= atoi(attr_values[i]);
			else
				CHECK(FALSE);
		}

		if(version < 7)
			FatalError(_("Early GuiXml version detected: resave in Crow 2.7.0 or later"));

		if(version > GUILOADER_VERSION_GUIXML || !StrEqual(nspace, "GTK+"))
			FatalError(_("Unsupported GuiXml version or namespace"));

		priv->model= node;
		priv->version= version;
		priv->stack= g_slist_prepend(priv->stack, node);

	} else {

		int i;
		for(i=0; attr_names[i]; ++i) {
			if(StrEqual(attr_names[i], "name"))
				node->name= AddSharedStr(user_data, ShortStrNew(attr_values[i]));
			else if(StrEqual(attr_names[i], "type"))
				node->type= ShortStrNew(attr_values[i]);
			else if(StrEqual(attr_names[i], "meta"))
				node->meta= ShortStrNew(attr_values[i]);
			else
				CHECK(FALSE);
		}

		if(StrEqual(element, "entity"))
			node->role= nrEntity;
		else if(StrEqual(element, "scalar"))
			node->role= nrScalar;
		else if(StrEqual(element, "vector"))
			node->role= nrVector;
		else if(StrEqual(element, "link"))
			node->role= nrLink;
		else
			CHECK(FALSE);

		node->owner= priv->stack->data;
		EnsureNodeDomain(node->owner);
		g_queue_push_tail(node->owner->domain, node);
		priv->stack= g_slist_prepend(priv->stack, node);

		if(node->owner->role==nrVector) {
			char buf[64];
			g_snprintf(buf, sizeof(buf), "%d", g_queue_get_length(node->owner->domain) - 1);
			CHECK(!node->name);
			node->name= AddSharedStr(user_data, ShortStrNew(buf)); //???
		}

	}

}

void OnEndElement(
	GMarkupParseContext * context,
	const char * element,
	gpointer user_data,
	GError ** error
) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(user_data);
	priv->stack= g_slist_delete_link(priv->stack, priv->stack);
}

void OnText(
	GMarkupParseContext * context,
	const char * text,
	gsize length,
	gpointer user_data,
	GError ** error
) {
	GuiLoaderPrivate * priv;
	Node * node;
	if(length == 0)
		return;
	priv= GUI_LOADER_GET_PRIVATE(user_data);
	node= priv->stack->data;
	CHECK(!node->text);
	if(node->role==nrScalar) {
		int size= length;
		const char * str= priv->preview ? text : Translate(ShortStrGet(node->meta), text, &size);
		node->text= ShortStrNewLen(str, size);
	} else if(node->role==nrLink)
		node->text= ShortStrNewLen(text, length);
}

void OnPassthrough(
	GMarkupParseContext * context,
	const char * text,
	gsize length,
	gpointer user_data,
	GError ** error
) {
	if(g_str_has_prefix(text, "<![CDATA[\n") && g_str_has_suffix(text, "]]>"))
		OnText(context, text+10, length-10-3, user_data, error);
}

void LoadModel(GuiLoader * loader, const char * data, int size) {
	GMarkupParseContext * context;

	static GMarkupParser parser= {
	  OnStartElement,
	  OnEndElement,
	  OnText,
	  OnPassthrough,
	  NULL
	};

	GError * error= NULL;

	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	CHECK(!priv->strings);

	context= g_markup_parse_context_new(&parser, 0, loader, NULL);

	if(!g_markup_parse_context_parse(context, data, size, &error))
		FatalError(error->message);

	if(!g_markup_parse_context_end_parse(context, &error))
		FatalError(error->message);

	g_markup_parse_context_free(context);

	CHECK(!priv->stack);

	if(priv->version <= GUI_NEEDS_UPGRADE) {
		g_warning(_("Upgrading GuiXml v%d -> v%d"), priv->version, GUILOADER_VERSION_GUIXML);
		switch(priv->version) {

			case 7:
			case 8:
				Upgrade8(loader, priv->model);

			case 9:
				Upgrade9(loader, priv->model);

		}
	}

	ForEachNode(priv->model, BindProxy, NULL);
}

void ForEachNode(Node * node, ForEachNodeFunc func, GSList ** list) {
	if(node) {
		if(node->domain) {
			GList * elem;
			for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem))
				ForEachNode(elem->data, func, list);
		}
		func(node, list);
	}
}

void Upgrade8(GuiLoader * loader, Node * node) {
	ForEachNode(node, U8_Rename, NULL);
	ForEachNode(node, U8_Fix, NULL);
}

void U8_Rename(Node * node, GSList ** list) {
	if(ShortStrGet(node->type)) {
		if(g_str_has_prefix(ShortStrGet(node->type), "Gideon"))
			node->type= ShortStrReplace(node->type, 0, 6, "Crow");
		if(StrEqual(ShortStrGet(node->type), "CrowPackOptions"))
			node->text= ShortStrReplace(node->text, 0, 6, "CROW");
	}
}

void U8_Fix(Node * node, GSList ** list) {
	if(ShortStrGet(node->type))
		if(
			StrEqual(ShortStrGet(node->type), "CrowBoxChild") && (
				StrEqual(ShortStrGet(node->owner->owner->type), "GtkHButtonBox") ||
				StrEqual(ShortStrGet(node->owner->owner->type), "GtkVButtonBox")
			)
		) {
			ShortStrUnref(node->type);
			node->type= ShortStrNew("CrowButtonBoxChild");
		}
}

void Upgrade9(GuiLoader * loader, Node * node) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	GSList * removals= NULL;
	GSList * tooltips= NULL;
	GSList * elem;
	ForEachNode(node, U9_GetRemovals, &removals);
	ForEachNode(node, U9_GetTooltips, &tooltips);
	for(elem= removals; elem; elem= g_slist_next(elem))
		U_RemoveNodes(elem->data);
	for(elem= tooltips; elem; elem= g_slist_next(elem)) {
		Node * ntooltips= elem->data;
		Node * ntips= FindNode(ntooltips, "tips");
		Node * nenabled= FindNode(ntooltips, "enabled");
		gboolean enabled= nenabled ? StrEqual(ShortStrGet(nenabled->text), "true") : TRUE;
		if(ntips) {
			GQueue * tips= ntips->domain;
			if(tips) {
				GList * elem;
				for(elem= g_queue_peek_head_link(tips); elem; elem= g_list_next(elem)) {
					Node * nwidget= FindNode(elem->data, "widget");
					Node * ntip= FindNode(elem->data, "tip-text");
					if(nwidget && ntip && (!ShortStrEmpty(ntip->text) || !ShortStrEmpty(ntip->meta))) {
						Node * widget= FindNode(priv->model, ShortStrGet(nwidget->text));
						if(widget && !FindNode(widget, "tooltip")) {
							Node * ntooltip= NewNode();
							ntooltip->role= nrScalar;
							ntooltip->name= ShortStrNew("tooltip");
							ntooltip->type= ShortStrNew("string");
							if(ntip->meta) {
								ntooltip->meta= ntip->meta;
								ShortStrRef(ntip->meta);
							}
							if(ntip->text) {
								ShortStrRef(ntip->text);
								ntooltip->text= ntip->text;
							}
							ntooltip->owner= widget;
							EnsureNodeDomain(widget);
							g_queue_push_tail(widget->domain, ntooltip);
							if(!enabled) {
								Node * nmode= NewNode();
								nmode->role= nrScalar;
								nmode->name= ShortStrNew("tooltip-mode");
								nmode->type= ShortStrNew("CrowTooltipMode");
								nmode->text= ShortStrNew("CROW_TOOLTIP_NONE");
								nmode->owner= widget;
								g_queue_push_tail(widget->domain, nmode);
							}
						}
					}
				}
			}
		}
		U_RemoveNodes(ntooltips);
	}
	g_slist_free(removals);
	g_slist_free(tooltips);
}

void U_RemoveNodes(Node * node) {
	GList * elem;
	for(elem= g_queue_peek_head_link(node->owner->domain); elem; elem= g_list_next(elem))
		if(elem->data == node) {
			g_queue_delete_link(node->owner->domain, elem);
			ForEachNode(node, FreeNode, NULL);
			return;
		}
	CHECK(FALSE);
}

void U9_GetRemovals(Node * node, GSList ** list) {
	if(StrEqual(ShortStrGet(node->name), "window-placement"))
		*list = g_slist_prepend(*list, node);
	if(StrEqual(ShortStrGet(node->name), "focus-child"))
		*list = g_slist_prepend(*list, node);
	if(StrEqual(ShortStrGet(node->name), "focus-chain")) {
		GType container= g_type_from_name(ShortStrGet(node->owner->type));
		if(g_type_is_a(container, GTK_TYPE_BIN))
			*list = g_slist_prepend(*list, node);
	}
}

void U9_GetTooltips(Node * node, GSList ** list) {
	if(StrEqual(ShortStrGet(node->type), "GtkTooltips"))
		*list = g_slist_prepend(*list, node);
}

void BindProxy(Node * node, GSList ** list) {
	if(node->role==nrEntity) {
		GuiLoaderClass * meta= GUI_LOADER_CLASS(g_type_class_ref(GUI_TYPE_LOADER));
		Proxy * proxy= g_hash_table_lookup(meta->init->proxies, ShortStrGet(node->type));
		CHECK(proxy && !node->proxy);
		node->proxy= proxy;
		g_type_class_unref(meta);
	}
}

Node * NewNode() {
	Node * node= g_slice_new(Node);
	node->role= nrNone;
	node->name= NULL;
	node->type= NULL;
	node->meta= NULL;
	node->text= NULL;
	node->owner= NULL;
	node->domain= NULL;
	node->proxy= NULL;
	node->object= NULL;
	return node;
}

void EnsureNodeDomain(Node * node) {
	if(!node->domain)
		node->domain= g_queue_new();
}

void FreeModel(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	CHECK(priv->stack==NULL);
	ForEachNode(priv->model, FreeNode, NULL);
	priv->model= NULL;
	priv->version= 0;
}

void FreeNode(Node * node, GSList ** list) {
	if(node) {
		if(node->name)
			ShortStrUnref(node->name);
		if(node->type)
			ShortStrUnref(node->type);
		if(node->meta)
			ShortStrUnref(node->meta);
		if(node->text)
			ShortStrUnref(node->text);
		if(node->domain)
			g_queue_free(node->domain);
		g_slice_free(Node, node);
	}
}

Node * FindNode(Node * node, const char * name) {
	if(node->domain) {
		GList * elem;
		for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
			Node * N= elem->data;
			if(StrEqual(name, ShortStrGet(N->name)))
				return N;
		}
	}
	return NULL;
}

ShortStr * AddSharedStr(GuiLoader * loader, ShortStr * string) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	priv->strings= g_slist_prepend(priv->strings, string);
	ShortStrRef(string);
	return string;
}

void LoadGui(GuiLoader * loader) {
	Node * root;
	int updated_count= 0;
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	PropertyFuncData data;
	GSList * elem;
	GList * i1, * i2;

	CHECK(!priv->delayed);
	root= priv->model;
	EnsureNodeDomain(root);
	g_queue_sort(root->domain, RootOrder, NULL);

	i1= g_queue_peek_head_link(root->domain);
	i2= g_queue_peek_head_link(root->domain);
	while(i2) {
		GList * in= g_list_next(i2);
		int p2= ((Node *)i2->data)->proxy->priority;
		int pn= !in ? 100 : ((Node *)in->data)->proxy->priority;
		if(pn > p2) {
			GList * ii;
			for(ii= i1; ; ii= g_list_next(ii)) {
				Node * node= ii->data;
				++updated_count;
				MakeObject(loader, node, NULL);
				g_hash_table_insert(priv->objects, ShortStrGet(node->name), node->object);
				if(ii==i2)
					break;
			}
			for(ii= i1; ; ii= g_list_next(ii)) {
				LoadObject(loader, ii->data);
				if(ii==i2)
					break;
			}
			i1= in;
		}
		i2= in;
	}

	CHECK(updated_count==SIGNED(g_queue_get_length(root->domain)));

	data.loader= loader;
	for(elem= priv->delayed; elem; elem= g_slist_next(elem)) {
		DelayedAssignment * assignment= elem->data;
		data.object= assignment->object;
		data.name= assignment->property->name;
		data.node= assignment->node;
		assignment->property->assign(&data);
		g_slice_free(DelayedAssignment, assignment);
	}
	g_slist_free(priv->delayed);
	priv->delayed= NULL;

}

int RootOrder(gconstpointer pa, gconstpointer pb, gpointer usedata) {
	int a= ((Node *)pa)->proxy->priority;
	int b= ((Node *)pb)->proxy->priority;
	return a - b;
}

void FreeGui(GuiLoader * loader) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	GSList * elem;
	g_hash_table_foreach_remove(priv->objects, RemoveObject, loader);
	for(elem= priv->nested; elem; elem= g_slist_next(elem))
		g_object_unref(elem->data);
	g_slist_free(priv->nested);
	priv->nested= NULL;
	g_hash_table_foreach_remove(priv->connectors, RemoveConnector, loader);
	for(elem= priv->strings; elem; elem= g_slist_next(elem))
		ShortStrUnref(elem->data);
	g_slist_free(priv->strings);
	priv->strings= NULL;
}

gboolean RemoveObject(gpointer key, gpointer value, gpointer userdata) {
	GObject * object= G_OBJECT(value);
	if(GTK_IS_WINDOW(object)) {
		gboolean take= GUI_LOADER_GET_PRIVATE(userdata)->take;
		gboolean bound= g_object_get_data(object, "GuiLoaderBoundWindow") != NULL;
		if(!take || !bound)
			gtk_object_destroy(GTK_OBJECT(object));
	}
	g_object_unref(object);
	return TRUE;
}

gboolean RemoveConnector(gpointer key, gpointer value, gpointer userdata) {
	GQueue * queue= value;
	GList * elem;
	for(elem= g_queue_peek_head_link(queue); elem; elem= g_list_next(elem)) {
		GuiConnector * connector= elem->data;
		ShortStrUnref(connector->emitter->signame);
		ShortStrUnref(connector->emitter->mapname);
		g_slice_free(CrowEmitter, connector->emitter);
		g_slice_free(GuiConnector, connector);
	}
	g_queue_free(queue);
	return TRUE;
}

void MakeObject(GuiLoader * loader, Node * node, GObject * old) {
	GObject * object;
	GList * elem;
	CHECK(node->proxy && !node->object);

	object= old;
	if(!object) {
		if(node->proxy->createInstance)
			object= node->proxy->createInstance(loader, node);
		else
			object= g_object_new(node->proxy->type, NULL);
	}

	if(!old) {
	  GuiLoaderClass * meta= GUI_LOADER_GET_CLASS(loader);
		MonitorObject(meta->init, object);
		if(g_object_is_floating(object))
			g_object_ref_sink(object);
		else if(GTK_IS_WINDOW(object))
			g_object_ref(object);
	} else {
		CHECK(!g_object_is_floating(object));
		CHECK(!GTK_IS_WINDOW(object));
	}

	for(elem= g_queue_peek_head_link(node->proxy->initInstanceList); elem; elem= g_list_next(elem)) {
		InstanceInitFunc initInstance= elem->data;
		initInstance(object, node);
	}

	node->object= object;
}

void LoadObject(GuiLoader * loader, Node * node) {
	PropertyFuncData data;
	GList * elem;
	CHECK(node->proxy && node->object);
	data.loader= loader;
	data.object= node->object;
	for(elem= g_queue_peek_head_link(node->proxy->properties); elem; elem= g_list_next(elem)) {
		Property * property= elem->data;
		Node * pnode= FindNode(node, property->name);
		if(pnode) {
			if(!property->lazy) {
				data.name= property->name;
				data.node= pnode;
				property->assign(&data);
			} else
				AddDelayedAssignment(loader, node->object, property, pnode);
		}
	}
}

void AddConnector(GuiLoader * loader, GObject * object, CrowEmitter * emitter) {
	GuiConnector * connector;
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	GQueue * queue= g_hash_table_lookup(priv->connectors, ShortStrGet(emitter->mapname));
	if(!queue) {
		queue= g_queue_new();
		g_hash_table_insert(priv->connectors, ShortStrGet(emitter->mapname), queue);
	}
	connector= g_slice_new(GuiConnector);
	connector->object= object;
	connector->emitter= emitter;
	g_queue_push_tail(queue, connector);
}

void AddDelayedAssignment(GuiLoader * loader, GObject * object, Property * property, Node * node) {
	GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
	DelayedAssignment * assignment= g_slice_new(DelayedAssignment);
	assignment->object= object;
	assignment->property= property;
	assignment->node= node;
	priv->delayed= g_slist_prepend(priv->delayed, assignment);
}

GObject * FindObject(GuiLoader * loader, const char * name) {
	GuiLoaderPrivate * priv;
	CHECK(name);
	priv= GUI_LOADER_GET_PRIVATE(loader);
	return g_hash_table_lookup(priv->objects, name);
}

void InitGettext(void) {
	#if ENABLE_NLS
		static gboolean done= FALSE;
		if(!done) {
			bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
			bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
			done= TRUE;
		}
	#endif
}

GuiLoaderInit * NewInit(void) {
	GuiLoaderInit * init= g_slice_new(GuiLoaderInit);
	CHECK(GTK_MAJOR_VERSION*1000+GTK_MINOR_VERSION <= gtk_major_version*1000+gtk_minor_version);
	init->refcounter= 1;
	init->proxies= g_hash_table_new(g_str_hash, g_str_equal);
	init->counters= NULL;
	init->MQ= g_quark_from_static_string("GuiLoaderQuark");
	RegisterProxies(init);
	return init;
}

void FreeInit(GuiLoaderInit * init) {
	CHECK(init->refcounter==0 && !init->counters);
	g_hash_table_foreach_remove(init->proxies, RemoveProxy, NULL);
	g_slice_free(GuiLoaderInit, init);
}

void EnterCounter(GuiLoaderInit * init, const char * message) {
	Counter * counter= g_slice_new(Counter);
	counter->value= 0;
	counter->message= message;
	init->counters= g_slist_prepend(init->counters, counter);
}

void LeaveCounter(GuiLoaderInit * init) {
	Counter * counter;
	CHECK(init->counters);
	counter= init->counters->data;
	if(counter->value != 0)
		g_warning(_("%d leak(s) in %s"), counter->value, counter->message ? counter->message : "<unknown>");
	g_slice_free(Counter, counter);
	init->counters= g_slist_delete_link(init->counters, init->counters);
}

void MonitorObject(GuiLoaderInit * init, GObject * object) {
	++((Counter *)init->counters->data)->value;
	g_object_set_qdata_full(object, init->MQ, init, (GDestroyNotify)ObjectDestroyed);
}

void ObjectDestroyed(GuiLoaderInit * init) {
	--((Counter *)init->counters->data)->value;
}

gboolean RemoveProxy(gpointer key, gpointer value, gpointer userdata) {
	Proxy * proxy= value;
	GList * elem;
	for(elem= g_queue_peek_head_link(proxy->properties); elem; elem= g_list_next(elem))
		g_slice_free(Property, elem->data);
	g_queue_free(proxy->properties);
	g_queue_free(proxy->initInstanceList);
	g_slice_free(Proxy, proxy);
	return TRUE;
}

void AddProxy(GuiLoaderInit * init, GType type, ProxyCreateFunc create, int priority) {
	Proxy * proxy= g_slice_new(Proxy);
	proxy->type= type;
	proxy->properties= g_queue_new();
	proxy->createInstance= NULL;
	proxy->initInstanceList= g_queue_new();
	proxy->bases= 0;
	proxy->priority= priority;
	create(proxy);
	g_hash_table_insert(init->proxies, (gpointer)g_type_name(type), proxy);
}

Property * AddProperty(Proxy * proxy, const char * name, PropertyFunc func) {
	Property * property;
	#ifdef GUILOADER_SANDBOX
		CHECK(!FindProperty(proxy, name));
	#endif
	property= g_slice_new(Property);
	property->name= name;
	property->assign= func;
	property->lazy= FALSE;
	g_queue_push_tail(proxy->properties, property);
	return property;
}

Property * FindProperty(Proxy * proxy, const char * name) {
	GList * elem;
	for(elem= g_queue_peek_head_link(proxy->properties); elem; elem= g_list_next(elem)) {
		Property * property= elem->data;
		if(StrEqual(name, property->name))
			return property;
	}
	return NULL;
}

void SetPropertyFunc(Proxy * proxy, const char * name, PropertyFunc func) {
	Property * property= FindProperty(proxy, name);
	property->assign= func;
}

void SetPropertyLazy(Proxy * proxy, const char * name, gboolean lazy) {
	Property * property= FindProperty(proxy, name);
	property->lazy= lazy;
}

void AddInitInstance(Proxy * proxy, InstanceInitFunc func) {
	g_queue_push_tail(proxy->initInstanceList, func);
}

void SetCreateInstance(Proxy * proxy, InstanceCreateFunc func) {
	proxy->createInstance= func;
}

void CheckNodeScalar(Node * node) {
	CHECK(node->role==nrScalar);
	CHECK(node->type);
	CHECK(!node->domain);
}

void CheckNodeVector(Node * node) {
	CHECK(node->role==nrVector);
	CHECK(!node->type);
	CHECK(node->domain && !g_queue_is_empty(node->domain));
}

void CheckNodeObject(Node * node) {
	CHECK(
		(node->role==nrEntity && node->type) ||
		(node->role==nrLink && !node->type && !node->domain)
	);
}

int ModelGetInt(Node * node) {
	char * endptr, * str;
	int var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= strtol(str, &endptr, 10);
	CHECK(*str!=0 && *endptr==0);
	return var;
}

unsigned ModelGetUnsigned(Node * node) {
	char * endptr, * str;
	unsigned var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= strtoul(str, &endptr, 10);
	CHECK(*str!=0 && *endptr==0);
	return var;
}

gboolean ModelGetBool(Node * node) {
	gboolean var;
	CheckNodeScalar(node);
	var= BoolFromString(ShortStrGet(node->text));
	return var;
}

gfloat ModelGetFloat(Node * node) {
	char * endptr, * str;
	gfloat var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= g_ascii_strtod(str, &endptr);
	CHECK(*str!=0 && *endptr==0);
	return var;
}

gdouble ModelGetDouble(Node * node) {
	char * endptr;
	char * str;
	gdouble var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= g_ascii_strtod(str, &endptr);
	CHECK(*str!=0 && *endptr==0);
	return var;
}

char * ModelGetString(Node * node) {
	char * str;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	return str ? str : "";
}

gunichar ModelGetChar(Node * node) {
	char * str;
	gunichar var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	CHECK(g_utf8_strlen(str, -1) == 1);
	var= g_utf8_get_char(str);
	return var;
}

int ModelGetEnum(Node * node) {
	GType type;
	GEnumClass * meta;
	GEnumValue * evalue;
	int var;
	CheckNodeScalar(node);
	type= g_type_from_name(ShortStrGet(node->type));
	CHECK(G_TYPE_IS_ENUM(type));
	meta= G_ENUM_CLASS(g_type_class_ref(type));
	evalue= g_enum_get_value_by_name(meta, ShortStrGet(node->text));
	CHECK(evalue);
	var= evalue->value;
	g_type_class_unref(meta);
	return var;
}

CrowPackOptions ModelGetPackOptions(Node * node) {
	char * str;
	CrowPackOptions var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= CROW_PACK_SHRINK;
	if(StrEqual(str, "CROW_PACK_SHRINK"))
		var= CROW_PACK_SHRINK;
	else if(StrEqual(str, "CROW_PACK_EXPAND_PADDING"))
		var= CROW_PACK_EXPAND_PADDING;
	else if(StrEqual(str, "CROW_PACK_EXPAND_WIDGET"))
		var= CROW_PACK_EXPAND_WIDGET;
	else
		CHECK(FALSE);
	return var;
}

CrowTooltipMode ModelGetTooltipMode(Node * node) {
	char * str;
	CrowTooltipMode var;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	var= CROW_TOOLTIP_TEXT;
	if(StrEqual(str, "CROW_TOOLTIP_MARKUP"))
		var= CROW_TOOLTIP_MARKUP;
	else if(StrEqual(str, "CROW_TOOLTIP_TEXT_OFF"))
		var= CROW_TOOLTIP_TEXT_OFF;
	else if(StrEqual(str, "CROW_TOOLTIP_MARKUP_OFF"))
		var= CROW_TOOLTIP_MARKUP_OFF;
	else if(StrEqual(str, "CROW_TOOLTIP_TEXT"))
		var= CROW_TOOLTIP_TEXT;
	else
		CHECK(FALSE);
	return var;
}

CrowPoint ModelGetPoint(Node * node) {
	char * str;
	CrowPoint point;
	int n=0;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	sscanf(str, "%d,%d%n", &point.x, &point.y, &n);
	CHECK(*str!=0 && *(str + n)==0);
	return point;
}

GtkBorder ModelGetGtkBorder(Node * node) {
	char * str;
	GtkBorder border;
	int n=0;
	CheckNodeScalar(node);
	str= ShortStrGet(node->text);
	sscanf(str, "%d,%d,%d,%d%n", &border.left, &border.right, &border.top, &border.bottom, &n);
	CHECK(*str!=0 && *(str + n)==0);
	return border;
}

GdkColor ModelGetColor(Node * node) {
	GdkColor color= { 0, };
	int result;
	CheckNodeScalar(node);
	result= gdk_color_parse(ShortStrGet(node->text), &color);
	CHECK(result==TRUE);
	return color;
}

CrowEmitter * ModelGetEmitter(Node * node) {
	int count= 3;
	pconstchar * ptrs;
	CrowEmitter * emitter;
	CheckNodeScalar(node);
	ptrs= SplitString(ShortStrGet(node->text), ',', &count);
	CHECK(count==3);
	emitter= g_slice_new(CrowEmitter);
	emitter->signame= ShortStrNewLen(ptrs[0], SplitLength(ptrs, 0));
	emitter->mapname= ShortStrNewLen(ptrs[1], SplitLength(ptrs, 1));
	emitter->after= BoolFromString(ptrs[2]);
	FreeSplitPtrs(ptrs, count);
	return emitter;
}

GObject * ModelGetObject(GuiLoader * loader, Node * node, GObject * old) {
	CheckNodeObject(node);
	if(node->role==nrLink) {
		const char * target= ShortStrGet(node->text);
		return target ? FindObject(loader, target) : NULL;
	} else {
		GuiLoaderPrivate * priv= GUI_LOADER_GET_PRIVATE(loader);
		CHECK(node->owner != priv->model);
		MakeObject(loader, node, old);
		if(!old)
			priv->nested= g_slist_prepend(priv->nested, node->object);
		LoadObject(loader, node);
		return node->object;
	}
}

PtrArray * ModelGetObjectVector(GuiLoader * loader, Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		GObject * object= ModelGetObject(loader, item, NULL);
		result->ptrs[i++]= object;
	}
	return result;
}

PtrArray * ModelGetStringVector(Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		char * string= ModelGetString(item);
		result->ptrs[i++]= string;
	}
	return result;
}

PtrArray * ModelGetEmitterVector(Node * node) {
	PtrArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= PtrArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		CrowEmitter * emitter= ModelGetEmitter(item);
		result->ptrs[i++]=  emitter;
	}
	return result;
}

IntArray * ModelGetEnumVector(Node * node) {
	IntArray * result;
	GList * elem;
	int i= 0;
	CheckNodeVector(node);
	result= IntArrayNew(g_queue_get_length(node->domain));
	for(elem= g_queue_peek_head_link(node->domain); elem; elem= g_list_next(elem)) {
		Node * item= elem->data;
		int value= ModelGetEnum(item);
		result->data[i++]= value;
	}
	return result;
}

void PropertyInt(PropertyFuncData * data) {
	int var= ModelGetInt(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_INT);
	g_value_set_int(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyUnsigned(PropertyFuncData * data) {
	unsigned var= ModelGetUnsigned(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_UINT);
	g_value_set_uint(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyBool(PropertyFuncData * data) {
	gboolean var= ModelGetBool(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_BOOLEAN);
	g_value_set_boolean(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyFloat(PropertyFuncData * data) {
	gfloat var= ModelGetFloat(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_FLOAT);
	g_value_set_float(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyDouble(PropertyFuncData * data) {
	gdouble var= ModelGetDouble(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_DOUBLE);
	g_value_set_double(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyString(PropertyFuncData * data) {
	const char * var= ModelGetString(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_STRING);
	g_value_set_static_string(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyChar(PropertyFuncData * data) {
	gunichar var= ModelGetChar(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_UINT);
	g_value_set_uint(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyEnum(PropertyFuncData * data) {
	int var= ModelGetEnum(data->node);
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_INT);
	g_value_set_int(&value, var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void PropertyObject(PropertyFuncData * data) {
	GObject * old, * var;
	GValue value= { 0, };
	g_value_init(&value, G_TYPE_OBJECT);
	g_object_get_property(data->object, data->name, &value);
	old= g_value_get_object(&value);
	var= ModelGetObject(data->loader, data->node, old);
	CHECK(old ? old==var : var!=NULL);
	if(var!=old) {
		g_value_set_object(&value, var);
		g_object_set_property(data->object, data->name, &value);
	}
	g_value_unset(&value);
}

void PropertyColor(PropertyFuncData * data) {
	GdkColor var= ModelGetColor(data->node);
	GValue value= { 0, };
	g_value_init(&value, GDK_TYPE_COLOR);
	g_value_set_static_boxed(&value, &var);
	g_object_set_property(data->object, data->name, &value);
	g_value_unset(&value);
}

void DrawGrid(GdkDrawable * window, GtkStyle * style, gboolean squares) {
	GdkGC * gc= gdk_gc_new(window);
	int W, H, GS, GW, GH, xx, yy, GX, GY, x, y;
	gdk_drawable_get_size(window, &W, &H);
	gdk_gc_set_foreground(gc, &style->dark[GTK_STATE_NORMAL]);
	gdk_draw_rectangle(window, gc, FALSE, 1, 1, W-3, H-3);
	if(squares)
		gdk_gc_set_foreground(gc, &style->mid[GTK_STATE_NORMAL]);
	GS= 9;
	GW= W-4;
	GH= H-4;
	xx= GW/GS;
	yy= GH/GS;
	GX= 2 + (GW-xx*GS)/2;
	GY= 2 + (GH-yy*GS)/2;
	if(!squares) {
		for(x=0; x<xx; ++x)
			for(y=0; y<yy; ++y)
				gdk_draw_point(window, gc, GX+x*GS+GS/2, GY+y*GS+GS/2);
	} else {
		for(x=0; x<xx-1; ++x)
			for(y=0; y<yy-1; ++y)
				if((x+y) % 2 == 0)
					gdk_draw_rectangle(window, gc, TRUE, GX+x*GS+GS/2, GY+y*GS+GS/2, GS, GS);
	}
	g_object_unref(gc);
}

GdkPixbuf * LoadImage(const char * name) {
	GdkPixbuf * pixbuf= NULL;
	if(name && strlen(name) > 0) {
		GtkIconInfo * info;
		GtkIconTheme * theme= gtk_icon_theme_get_default();
		CHECK(theme);
		info= gtk_icon_theme_lookup_icon(theme, name, -1, 0);
		if(info) {
			const char * filename= gtk_icon_info_get_filename(info);
			if(filename)
				pixbuf= gdk_pixbuf_new_from_file(filename, NULL);
		}
		if(!pixbuf)
			pixbuf= gtk_icon_theme_load_icon(theme, "gtk-missing-image", 16, GTK_ICON_LOOKUP_FORCE_SIZE, NULL);
	}
	return pixbuf;
}
