/*
 *  GuiLoader, a GuiXml loader library
 *  Copyright (c) 2006-2010 Maxim Udushlivy
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  The software is provided "as is", without warranty of any kind,
 *  express or implied, including but not limited to the warranties of
 *  merchantability, fitness for a particular purpose and
 *  noninfringement. In no event shall the authors or copyright
 *  holders be liable for any claim, damages or other liability,
 *  whether in an action of contract, tort or otherwise, arising from,
 *  out of or in connection with the software or the use or other
 *  dealings in the software.
 */

#ifndef GUILOADER_GUILOADERIMPL_H
#define GUILOADER_GUILOADERIMPL_H

#define GUILOADER_LIBRARY

#ifdef GUILOADER_SANDBOX
	#define G_DISABLE_DEPRECATED
	#define PANGO_DISABLE_DEPRECATED
	#define ATK_DISABLE_DEPRECATED
	#define GDK_DISABLE_DEPRECATED
	#define GDK_PIXBUF_DISABLE_DEPRECATED
	#define GTK_DISABLE_DEPRECATED
	//#define GSEAL_ENABLE
#endif

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "guiloader.h"

#include <atk/atk-enum-types.h>

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#define GETTEXT_PACKAGE "guiloader"
#include <glib/gi18n-lib.h>

#define GUILOADER_VERSION_GUIXML (12)
#define GUI_NEEDS_UPGRADE (9)

#ifdef G_LOG_DOMAIN
	#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "GuiLoader"

#define GUI_LOADER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), GUI_TYPE_LOADER, GuiLoaderPrivate))
#define CHECK(expr) ((expr) ? NoOp() : (CheckFailed(#expr, __FILE__, __LINE__), NoOp()))
#define SIGNED(value) ((int)value)

#define GUILOADER_VIRTUAL_BASE(proxy, vb)  \
  if((proxy->bases & vb) == vb)            \
    return;                                \
  else                                     \
    proxy->bases |= vb;

typedef enum {
	vbGObject=        1,
	vbGtkWidget=      2,
	vbGtkContainer=   4
} VirtualBases;

typedef enum {
	CROW_PACK_SHRINK,
	CROW_PACK_EXPAND_PADDING,
	CROW_PACK_EXPAND_WIDGET
} CrowPackOptions;

typedef enum {
	CROW_TOOLTIP_TEXT,
	CROW_TOOLTIP_TEXT_OFF,
	CROW_TOOLTIP_MARKUP,
	CROW_TOOLTIP_MARKUP_OFF
} CrowTooltipMode;

typedef enum { nrNone, nrEntity, nrScalar, nrVector, nrLink } NodeRole;

typedef struct _GuiLoaderPrivate GuiLoaderPrivate;
typedef struct _Counter Counter;
typedef struct _Node Node;
typedef struct _PropertyFuncData PropertyFuncData;
typedef struct _Property Property;
typedef struct _Proxy Proxy;
typedef struct _CrowPoint CrowPoint;
typedef struct _DelayedAssignment DelayedAssignment;
typedef struct _PtrArray PtrArray;
typedef struct _IntArray IntArray;
typedef struct _CrowEmitter CrowEmitter;
typedef struct _ShortStr ShortStr;
typedef struct _CustomConstructor CustomConstructor;

typedef GObject * (*InstanceCreateFunc)(GuiLoader *, Node *);
typedef void (*InstanceInitFunc)(GObject *, Node *);
typedef void (*PropertyFunc)(PropertyFuncData *);
typedef void (*ProxyCreateFunc)(Proxy *);
typedef void (*ForEachNodeFunc)(Node *, GSList ** list);

typedef const char * pconstchar;

struct _GuiLoaderPrivate {
	gboolean preview;
	gboolean take;
	GSList * stack;
	Node * model;
	int version;
	GHashTable * objects;
	GSList * nested;
	GHashTable * connectors;
	GSList * delayed;
	GSList * strings;
	GHashTable * constructors;
};

struct _Counter {
	int value;
	const char * message;
};

struct _GuiLoaderInit {
	int refcounter;
	GHashTable * proxies;
	GSList * counters;
	GQuark MQ;
};

struct _Node {
	NodeRole role;
	ShortStr * name;
	ShortStr * type;
	ShortStr * meta;
	ShortStr * text;
	Node * owner;
	GQueue * domain;
	Proxy * proxy;
	GObject * object;
};

struct _PropertyFuncData {
	GuiLoader * loader;
	GObject * object;
	const char * name;
	Node * node;
};

struct _Property {
	const char * name;
	PropertyFunc assign;
	gboolean lazy;
};

struct _Proxy {
	GType type;
	GQueue * properties;
	InstanceCreateFunc createInstance;
	GQueue * initInstanceList;
	VirtualBases bases;
	int priority;
};

struct _CrowPoint {
	int x;
	int y;
};

struct _DelayedAssignment {
	GObject * object;
	Property * property;
	Node * node;
};

struct _PtrArray {
	int size;
	gpointer ptrs[1];
};

struct _IntArray {
	int size;
	int data[1];
};

struct _CrowEmitter {
	ShortStr * signame;
	ShortStr * mapname;
	gboolean after;
};

struct _GuiConnector {
	GObject * object;
	CrowEmitter * emitter;
};

struct _ShortStr {
	union {
		struct {
			guint32 len : 16;
			guint32 refcounter : 8;
			guint32 ch : 8;
		};
		char buf[4];
	};
};

struct _CustomConstructor {
	GuiCustomConstructorFunc func;
	gpointer data;
	GDestroyNotify destroy;
};

void gui_loader_class_init(GuiLoaderClass * meta);
void gui_loader_init(GuiLoader * loader);
void gui_loader_finalize(GObject * object);

void CheckFailed(const char * expr, const char * file, int line);
void FatalError(const char * message);
int NoOp(void);

ShortStr * ShortStrNew(const char * str);
ShortStr * ShortStrNewLen(const char * str, int length);
ShortStr * ShortStrAlloc(int length);
char * ShortStrGet(ShortStr * string);
void ShortStrRef(ShortStr * string);
void ShortStrUnref(ShortStr * string);
ShortStr * ShortStrReplace(ShortStr * string, int pos, int len, const char * repstr);
guint ShortStrHash(gconstpointer string);
gboolean ShortStrEqual(gconstpointer s1, gconstpointer s2);
gboolean ShortStrEmpty(ShortStr * string);

PtrArray * PtrArrayNew(int size);
void PtrArrayFree(PtrArray * array);
void PtrArrayNull(PtrArray * array);
IntArray * IntArrayNew(int size);
void IntArrayFree(IntArray * array);

GList * CreateListFromPtrArray(PtrArray * array);
pconstchar * SplitString(const char * str, char ch, int * maxcount);
int SplitLength(pconstchar * ptrs, int i);
void FreeSplitPtrs(pconstchar * ptrs, int count);
gboolean BoolFromString(const char * str);
const char * Translate(const char * meta, const char * str, int * size);
gboolean StrEqual(const char * s1, const char * s2);

void InitInstance(GuiLoader * loader);
void DoneInstance(GuiLoader * loader);

void CustomConstructorFree(CustomConstructor * constructor);

void LoadModel(GuiLoader * loader, const char * data, int size);
void ForEachNode(Node * node, ForEachNodeFunc func, GSList ** list);
void Upgrade8(GuiLoader * loader, Node * node);
void U8_Rename(Node * node, GSList ** list);
void U8_Fix(Node * node, GSList ** list);
void Upgrade9(GuiLoader * loader, Node * node);
void U_RemoveNodes(Node * node);
void U9_GetRemovals(Node * node, GSList ** list);
void U9_GetTooltips(Node * node, GSList ** list);
void BindProxy(Node * node, GSList ** list);
Node * NewNode();
void EnsureNodeDomain(Node * node);
void FreeModel(GuiLoader * loader);
void FreeNode(Node * node, GSList ** list);
Node * FindNode(Node * node, const char * name);
ShortStr * AddSharedStr(GuiLoader * loader, ShortStr * string);

void LoadGui(GuiLoader * loader);
int RootOrder(gconstpointer pa, gconstpointer pb, gpointer usedata);
void FreeGui(GuiLoader * loader);
gboolean RemoveObject(gpointer key, gpointer value, gpointer userdata);
gboolean RemoveConnector(gpointer key, gpointer value, gpointer userdata);
void MakeObject(GuiLoader * loader, Node * node, GObject * old);
void LoadObject(GuiLoader * loader, Node * node);
void AddConnector(GuiLoader * loader, GObject * object, CrowEmitter * emitter);
void AddDelayedAssignment(GuiLoader * loader, GObject * object, Property * property, Node * node);
GObject * FindObject(GuiLoader * loader, const char * name);

void InitGettext(void);
GuiLoaderInit * NewInit(void);
void FreeInit(GuiLoaderInit * init);
void EnterCounter(GuiLoaderInit * init, const char * message);
void LeaveCounter(GuiLoaderInit * init);
void MonitorObject(GuiLoaderInit * init, GObject * object);
void ObjectDestroyed(GuiLoaderInit * init);
gboolean RemoveProxy(gpointer key, gpointer value, gpointer userdata);
void AddProxy(GuiLoaderInit * init, GType type, ProxyCreateFunc create, int priority);
Property * AddProperty(Proxy * proxy, const char * name, PropertyFunc func);
Property * FindProperty(Proxy * proxy, const char * name);
void SetPropertyFunc(Proxy * proxy, const char * name, PropertyFunc func);
void SetPropertyLazy(Proxy * proxy, const char * name, gboolean lazy);
void AddInitInstance(Proxy * proxy, InstanceInitFunc func);
void SetCreateInstance(Proxy * proxy, InstanceCreateFunc func);

void CheckNodeScalar(Node * node);
void CheckNodeVector(Node * node);
void CheckNodeObject(Node * node);

int ModelGetInt(Node * node);
unsigned ModelGetUnsigned(Node * node);
gboolean ModelGetBool(Node * node);
gfloat ModelGetFloat(Node * node);
gdouble ModelGetDouble(Node * node);
char * ModelGetString(Node * node);
gunichar ModelGetChar(Node * node);
int ModelGetEnum(Node * node);
CrowPackOptions ModelGetPackOptions(Node * node);
CrowTooltipMode ModelGetTooltipMode(Node * node);
CrowPoint ModelGetPoint(Node * node);
GtkBorder ModelGetGtkBorder(Node * node);
GdkColor ModelGetColor(Node * node);
CrowEmitter * ModelGetEmitter(Node * node);
GObject * ModelGetObject(GuiLoader * loader, Node * node, GObject * old);
PtrArray * ModelGetObjectVector(GuiLoader * loader, Node * node);
PtrArray * ModelGetStringVector(Node * node);
PtrArray * ModelGetEmitterVector(Node * node);
IntArray * ModelGetEnumVector(Node * node);

void PropertyInt(PropertyFuncData * data);
void PropertyUnsigned(PropertyFuncData * data);
void PropertyBool(PropertyFuncData * data);
void PropertyFloat(PropertyFuncData * data);
void PropertyDouble(PropertyFuncData * data);
void PropertyString(PropertyFuncData * data);
void PropertyChar(PropertyFuncData * data);
void PropertyEnum(PropertyFuncData * data);
void PropertyObject(PropertyFuncData * data);
void PropertyColor(PropertyFuncData * data);

void DrawGrid(GdkDrawable * window, GtkStyle * style, gboolean squares);
GdkPixbuf * LoadImage(const char * name);

#endif
